package com.evasion.sam.jaas;

import com.evasion.sam.Config;
import com.evasion.sam.PasswordEncoder;
import com.evasion.sam.ejb.JNDIClient;
import com.evasion.sam.ejb.JaasEjb;
import java.util.Date;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;
import javax.transaction.*;

/**
 * Module d'authentification JAAS. @TODO implémenter la récupération en base de
 * données.
 *
 * @author sebastien
 */
public class EvasionEJBLoginModule implements LoginModule {

    /**
     * LOGGER.
     */
    private static final Logger LOGGER = Logger.getLogger(EvasionEJBLoginModule.class.getName());
    
    private Subject subject;

    private String username = null;

    private String password = null;

    private CallbackHandler callbackHandler;

    private boolean success = true;

    private JaasEjb loginEJB = null;
        JNDIClient ejbClient = null;
    @Override
    public void initialize(Subject subject, CallbackHandler callbackHandler, Map<String, ?> sharedState, Map<String, ?> options) {

        this.subject = subject;
        this.callbackHandler = callbackHandler;
        LOGGER.info("Init Evasion Login Module");
        String jndi = (String) options.get(Config.PARAM_EJB_JNDI);
        String providerUrl = (String) options.get(Config.PARAM_PROVIDER_URL);

        LOGGER.info("------- Properties ------------------");
        completePropertiesAndLog(Config.PARAM_EJB_JNDI, jndi);
        completePropertiesAndLog(Config.PARAM_PROVIDER_URL, providerUrl);

        ejbClient = new JNDIClient(providerUrl);
        loginEJB = (JaasEjb) ejbClient.lookup(jndi);

    }

    private void completePropertiesAndLog(String propertyName, String propertyValue) {
        LOGGER.log(Level.INFO, "{0} : {1}", new Object[]{propertyName, propertyValue});
        if (propertyValue == null || propertyValue.equals("")) {
            LOGGER.log(Level.SEVERE, "{0} can not be null", propertyName);
        }
    }

    @Override
    public boolean login() throws LoginException {
        LOGGER.info("Start Login");
        traitementPWD();

        String dbPassword = loginEJB.getPassword(username);
        if (dbPassword == null || !dbPassword.equals(password)) {
            throw new LoginException("Bad username or password  for username=" + username);
        }
        LOGGER.fine("Login succes");
        return true;
    }

    private boolean traitementPWD() throws LoginException {
        try {

            NameCallback nc = new NameCallback("UsrName");
            PasswordCallback pc = new PasswordCallback("Passwd", false);
            callbackHandler.handle(new Callback[]{nc, pc});
            username = nc.getName();
            char[] tmp = pc.getPassword();
            if (tmp != null) {
                password = new String(tmp);
            }
            if (password == null || password.isEmpty() || username == null || username.isEmpty()) {
                LOGGER.severe("User or password are null");
                throw new LoginException("Login Failed for user " + username + "!!!");
            }

            password = PasswordEncoder.encodePassword(username, password);
            pc.clearPassword();
        } catch (Exception ex) {
            success = false;
            LoginException le = new LoginException("Login Failed!!!");
            LOGGER.log(Level.SEVERE, "Login Failed with username: {0} and password: xxxxxx", username);
            le.initCause(ex);
            throw le;
        }
        LOGGER.log(Level.FINE, "Login with username: {0} and password: xxxxxxxxx", username);
        return true;
    }

    @Override
    public boolean commit() throws LoginException {
       UserTransaction tx = ejbClient.getTransaction();       
        LOGGER.fine("Commit");
        if (username != null && success) {
            EvasionPrincipal principal = loginEJB.getEvasionPrincipal(username);
            subject.getPrincipals().add(principal);
            EvasionGroup roles = loginEJB.getAllRoles(username);
            subject.getPrincipals().add(roles);
            if ((new Date().getTime() - principal.getLastLogin().getTime()) > 86400) {
                try {
                    tx.begin();
                    loginEJB.postLogin(username);
                    tx.commit();
                } catch (RollbackException ex) {
                    Logger.getLogger(EvasionEJBLoginModule.class.getName()).log(Level.SEVERE, null, ex);
                } catch (HeuristicMixedException ex) {
                    Logger.getLogger(EvasionEJBLoginModule.class.getName()).log(Level.SEVERE, null, ex);
                } catch (HeuristicRollbackException ex) {
                    Logger.getLogger(EvasionEJBLoginModule.class.getName()).log(Level.SEVERE, null, ex);
                } catch (SecurityException ex) {
                    Logger.getLogger(EvasionEJBLoginModule.class.getName()).log(Level.SEVERE, null, ex);
                } catch (IllegalStateException ex) {
                    Logger.getLogger(EvasionEJBLoginModule.class.getName()).log(Level.SEVERE, null, ex);
                } catch (NotSupportedException ex) {
                    Logger.getLogger(EvasionEJBLoginModule.class.getName()).log(Level.SEVERE, null, ex);
                } catch (SystemException ex) {
                    Logger.getLogger(EvasionEJBLoginModule.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
        }
        return true;
    }

    @Override
    public boolean abort() throws LoginException {
        username = null;
        password = null;
        return true;
    }

    @Override
    public boolean logout() throws LoginException {
        subject.getPrincipals().remove(new EvasionPrincipal(username));
        username = null;
        password = null;
        return true;
    }
}
